"""Add all types of functions with fractions.

Add functions related to the operation and conversion of fractions
in a string format, with specific classes depending on what needs
to be done.

Fractions():
    separate_numbers(fraction):
        Separate fraction into numerator and denominator.
    sign_checker(fraction): Check sign of maximum fraction.
    simplify(fraction): Simplify fraction to his minimum expresion.
    add_zero_checker(fraction1, fraction2):
        Check if in an addition of two fractions returns zero.
    sub_zero_checker(fraction1, fraction2):
        Check if in a substraction of two fractions returns zero.

ConvertFractions():
    to_decimal(fraction, number_decimal):
        Convert a fraction into his decimal with an specific amount of
        decimals.
    to_mixed_number(fraction): Convert a fraction into his mixed number.
    to_fraction(number): Convert a decimal into his fraction.

OperationFractions():
    addition(*fractions): Add 2 or more fractions.
    substraction(*fractions): Substract 2 or more fractions.
    multiplication(*fractions): Multiply 2 or more fractions.
    division(*fractions): Divide 2 or more fractions.
"""

from .core import Fractions
from .convert import ConvertFractions
from .operation import OperationFractions

__tittle__ = "real fractions"
__url__ = "https://github.com/LuckJMG/RealFractions"
__author__ = "LuckJMG"
__email__ = "lucas.mosquera13@gmail.com"
__license__ = "MIT license"
__copyright__ = "Copyright 2020, LuckJMG"
__version__ = "1.1.1"
__status__ = "Production"
__all__ = ["ConvertFractions", "OperationFractions"]



"""Operation module from real_fractions package.

Classes:
    OperationFractions: Operation class from real_fractions.
"""

from .core import Fractions
from .convert import ConvertFractions
from math import sqrt

class OperationFractions(Fractions):
    """Operation class from real_fractions.

    Methods:
        addition(*fractions): Add 2 or more fractions.
        substraction(*fractions): Substract 2 or more fractions.
        multiplication(*fractions): Multiply 2 or more fractions.
        division(*fractions): Divide 2 or more fractions.
    """

    @classmethod
    def addition(cls, *fractions):
        """Add 2 or more fractions.

        Returns:
            str: Result from the addition of the fractions.
        """

        # Discard invalid arguments
        for index, fraction in enumerate(fractions):
            if not isinstance(fraction, str):  # Invalid argument types
                print("The argument {0} isn't a fraction".format(index))
                return None

            if cls.simplify(fraction) is None:  # Non fraction arguments
                print("The argument {0} isn't a fraction".format(index))
                return None

        tmp_fraction = "0/1"  # Neutral element

        try:
            for index, fraction in enumerate(fractions):
                if index != 0:
                    if cls.add_zero_checker(tmp_fraction, fraction):
                        # Separate numbers
                        numerator1, denominator1 = cls.separate_numbers(
                            tmp_fraction
                        )
                        numerator2, denominator2 = cls.separate_numbers(
                            fraction
                        )

                        # Calculate new fraction
                        new_denominator = str(denominator1 * denominator2)
                        new_numerator = str(
                            numerator1 * denominator2
                            + numerator2 * denominator1
                        )

                        # Make new fraction
                        tmp_fraction = new_numerator + "/" + new_denominator
                        tmp_fraction = cls.simplify(tmp_fraction)
                        numbers = cls.separate_numbers(tmp_fraction)
                        numerator = numbers[0]
                        denominator = numbers[1]

                        # Variations
                        if denominator == 1:
                            tmp_fraction = numerator

                        elif numerator == denominator:
                            tmp_fraction = 1

                    else: tmp_fraction = "0/1"  # The addition is zero

                else: tmp_fraction = fractions[0] # Start Point

            if tmp_fraction == "0/1":
                tmp_fraction = 0

            return tmp_fraction

        except ValueError:
            print("One of the arguments isn't a fraction")


    @classmethod
    def substraction(cls, *fractions):
        """Substract 2 or more fractions.

        Returns:
            str: Result from the substraction of the fractions.
        """

        # Discard invalid arguments
        for index, fraction in enumerate(fractions):
            if not isinstance(fraction, str):  # Invalid argument types
                print("The argument {0} isn't a fraction".format(index))
                return None

            if cls.simplify(fraction) is None:  # Non fraction arguments
                print("The argument {0} isn't a fraction".format(index))
                return None

        tmp_fraction = "0/1"  # Neutral element

        try:
            for index, fraction in enumerate(fractions):
                if index != 0:
                    if cls.sub_zero_checker(tmp_fraction, fraction):
                        # Separate numbers
                        numerator1, denominator1 = cls.separate_numbers(
                            tmp_fraction
                        )
                        numerator2, denominator2 = cls.separate_numbers(
                            fraction
                        )

                        # Calculate new fraction
                        new_denominator = str(denominator1 * denominator2)
                        new_numerator = str(
                            numerator1 * denominator2
                            - numerator2 * denominator1
                        )

                        # Make new fraction
                        tmp_fraction = new_numerator + "/" + new_denominator
                        tmp_fraction = cls.simplify(tmp_fraction)
                        numbers = cls.separate_numbers(tmp_fraction)
                        numerator = numbers[0]
                        denominator = numbers[1]

                        # Variations
                        if denominator == 1:
                            tmp_fraction = numerator

                        elif numerator == denominator:
                            tmp_fraction = 1

                    else: tmp_fraction = "0/1"  # The addition is zero

                else: tmp_fraction = fractions[0]  # Start point

            if tmp_fraction == "0/1":
                tmp_fraction = 0

            return tmp_fraction

        except ValueError:
            print("One of the arguments isn't a fraction")

    @classmethod
    def multiplication(cls, *fractions):
        """Multiply 2 or more fractions.

        Returns:
            str: Result from the multiplication of the fractions.
        """

        # Discard invalid arguments
        for index, fraction in enumerate(fractions):
            if not isinstance(fraction, str) and fraction != 0:
                # Invalid argument types
                print("The argument {0} isn't a fraction".format(index))
                return None

            if cls.simplify(fraction) is None:  # Non fraction arguments
                print("The argument {0} isn't a fraction".format(index))
                return None

        tmp_fraction = "1/1"  # Neutral element

        for index, fraction in enumerate(fractions):

            if fraction == 0 or tmp_fraction == 0:  # 0 * n = 0
                return 0

            if index != 0:
                # Multiplication
                numerator1, denominator1 = cls.separate_numbers(tmp_fraction)
                numerator2, denominator2 = cls.separate_numbers(fraction)
                new_denominator = denominator1 * denominator2
                new_numerator = numerator1 * numerator2

                # New fraction
                tmp_fraction = str(new_numerator) + "/" + str(new_denominator)
                tmp_fraction = cls.simplify(tmp_fraction)

            else: tmp_fraction = fractions[0]  # Start point

        return tmp_fraction

    @classmethod
    def division(cls, *fractions):
        """Divide 2 or more fractions.

        Returns:
            str: Result from the division of the fractions.
        """

        # Discard invalid arguments
        for index, fraction in enumerate(fractions):
            if not isinstance(fraction, str):  # Invalid argument types
                print("The argument {0} isn't a fraction".format(index))
                return None

            if cls.simplify(fraction) is None:  # Non fraction arguments
                print("The argument {0} isn't a fraction".format(index))
                return None

        tmp_fraction = "1/1"  # Neutral element

        for index, fraction in enumerate(fractions):
            if index != 0:
                # Division
                numerator1, denominator1 = cls.separate_numbers(tmp_fraction)
                numerator2, denominator2 = cls.separate_numbers(fraction)
                new_denominator = denominator1 * numerator2
                new_numerator = denominator2 * numerator1

                # New fraction
                tmp_fraction = str(new_numerator) + "/" + str(new_denominator)
                tmp_fraction = cls.simplify(tmp_fraction)

            else: tmp_fraction = fractions[0]  # Start point

        return tmp_fraction

    @classmethod
    def exponential(cls, fraction, exponent=2):
        """Raise the fraction to the indicated exponent, default exponent = 2.

        Returns:
            str: Result from the exponent of the fraction.
        """

        if not isinstance(fraction, str):  # Invalid first argument
            print("The first argument isn't a fraction")
            return None

        elif not isinstance(exponent, (int, float)):  # Invalid second argument
            print("The second argument isn't a number")
            return None

        if cls.simplify(fraction) is None:  # Non fraction arguments
            print("The first argument isn't a fraction")
            return None

        convert = ConvertFractions()

        if exponent < 0:
            # If the exponent is below 0, the fractions are reversed
            exponent = abs(exponent)
            numerator, denominator = cls.separate_numbers(fraction)
            fraction = str(denominator) + "/" + str(numerator)
            fraction = cls.simplify(fraction)

        # Exponent
        numerator, denominator = cls.separate_numbers(fraction)
        numerator = numerator ** exponent
        numerator = convert.to_fraction(numerator)
        denominator = denominator ** exponent
        denominator = convert.to_fraction(denominator)

        # New fraction
        fraction = cls.division(numerator, denominator)

        return fraction

    @classmethod
    def square_root(cls, fraction):
        """calculates the square root of the given fraction.

        Returns:
            str: Result from square root of the fraction.
        """

        if not isinstance(fraction, str):  # Invalid first argument
            print("The first argument isn't a fraction")
            return None

        if cls.simplify(fraction) is None:  # Non fraction arguments
            print("The first argument isn't a fraction")
            return None

        convert = ConvertFractions()

        # Square root
        numerator, denominator = cls.separate_numbers(fraction)
        numerator = sqrt(abs(numerator))
        numerator = convert.to_fraction(numerator)
        denominator = sqrt(abs(denominator))
        denominator = convert.to_fraction(denominator)

        # New fraction
        fraction = cls.division(numerator, denominator)

        return fraction



"""Core module from real_fractions package.

Classes:
    Fractions: Base real_fraction class.
"""


class Fractions:
    """Base class from real_fractions.

    Methods:
        separate_numbers(fraction):
            Separate fraction into numerator and denominator.
        sign_checker(fraction): Check sign of maximum fraction.
        simplify(fraction): Simplify fraction to his minimum expresion.
        add_zero_checker(fraction1, fraction2):
            Check if in an addition of two fractions returns zero.
        sub_zero_checker(fraction1, fraction2):
            Check if in a substraction of two fractions returns zero.
    """

    @classmethod
    def separate_numbers(cls, fraction):
        """Separate fraction into numerator and denominator.

        Args:
            fraction (str): Fraction to separate numbers

        Returns:

            int: Numerator of the fraction
            int: Denominator of the fraction
        """

        if not isinstance(fraction, str): # Discard invalid argument types
            print("The argument isn't a fraction")
            return None

        try:
            numbers = fraction.split("/")  # Separate numerator and denominator
            numerator = int(numbers[0])
            denominator = int(numbers[1])
            return numerator, denominator

        except ValueError:  # Discard invalid string arguments
            print("The argument isn't a fraction")

    @classmethod
    def sign_checker(cls, fraction):
        """Check sign of maximum fraction.

        Args:
            fraction (str): Fraction to check his sign

        Returns:
            str: Real fraction sign
        """

        if not isinstance(fraction, str):  # Discard invalid argument types
            print("The argument isn't a fraction")
            return None

        try:
            numerator, denominator = cls.separate_numbers(fraction)

            if numerator > 0 > denominator:  # Check if its negative
                fraction = str(-(numerator)) + "/" + str(abs(denominator))

            elif numerator < 0 and denominator < 0:  # Check if its positive
                fraction = str(abs(numerator)) + "/" + str(abs(denominator))

            elif denominator == 0:  # ZeroDivision Error
                print("Can't be divided by zero!")
                return None

            else:
                fraction = str(numerator) + "/" + str(denominator)

            return fraction

        except TypeError: # Discard invalid string arguments
            print("The argument isn't a fraction")

    @classmethod
    def simplify(cls, fraction):
        """Simplify fraction to his minimum expresion.

        Args:
            fraction (str): Fraction to simplify

        Returns:
            str: Simplified fraction
        """


        if not isinstance(fraction, str) and fraction != 0:  # Discard invalid argument types
            print("The argument isn't a fraction")
            return None
        elif fraction == 0:
            return fraction

        try:
            numerator, denominator = cls.separate_numbers(fraction)
            res = 0
            maximum = abs(max(numerator, denominator))
            minimum = abs(min(numerator, denominator))

            while minimum != 0:  # Calculate M.C.D
                res = minimum
                minimum = maximum % minimum
                maximum = res

            numerator = str(int(numerator / maximum))
            denominator = str(int(denominator / maximum))
            fraction = numerator + "/" + denominator
            fraction = cls.sign_checker(fraction)
            return fraction

        except TypeError:  # Discard invalid string argument
            print("The argument isn't a fraction")

    @classmethod
    def add_zero_checker(cls, fraction1, fraction2):
        """Check if in an addition of two fractions returns zero.

        Args:
            fraction1 (str): First fraction in the addition
            fraction2 (str): Second fraction in the addition

        Returns:
            bool: If false the addition of the fractions returns zero, if is
            true the addition don't return zero
        """

        # Discard invalid arguments type
        # First argument
        if not isinstance(fraction1, str) and fraction1 != 0:
            print("The first argument isn't a fraction")
            return None

        # Second argument
        elif not isinstance(fraction2, str) and fraction2 != 0:
            print("The second argument isn't a fraction")
            return None

        # Check the sign of the fractions
        if (
                "-" in fraction1
                and "-" not in fraction2
                and fraction1.replace("-", "") == fraction2
        ):  # First case in which the sum gives zero
            not_zero = False

        elif (
                "-" not in fraction1
                and "-" in fraction2
                and fraction1 == fraction2.replace("-", "")
        ):  # Second case in which the sum gives zero
            not_zero = False

        elif (
                cls.simplify(fraction1) is None
                or cls.simplify(fraction2) is None
        ):  # Invalid arguments
            return None

        else:  # Other cases when the addition doesn't return zero
            not_zero = True

        return not_zero

    @classmethod
    def sub_zero_checker(cls, fraction1, fraction2):
        """Check if in an substraction of two fractions returns zero.

        Args:
            fraction1 (str): First fraction in the substraction
            fraction2 (str): Second fraction in the substraction

        Returns:
            bool:
            If false the substraction of the fractions returns zero, if is
            true the substraction don't return zero
        """

        # Discard invalid arguments type
        # First argument
        if not isinstance(fraction1, str) and fraction1 != 0:
            print("The first argument isn't a fraction")
            return None

        # Second argument
        elif not isinstance(fraction2, str) and fraction2 != 0:
            print("The second argument isn't a fraction")
            return None

        # Check the sign of the fractions
        if (
                fraction1 == fraction2
        ):  # First case in which the substraction gives zero
            not_zero = False

        elif (
                cls.simplify(fraction1) is None
                or cls.simplify(fraction2) is None
        ):  # Invalid arguments
            return None

        else:  # Other cases when the substraction doesn't return zero
            not_zero = True

        return not_zero



"""Convert module from real_fractions package.

Classes:
    ConvertFractions: Convert class from real_fractions.
"""

from math import trunc
from .core import Fractions

class ConvertFractions(Fractions):
    """Convert class from real_fractions.

    Methods:
        to_decimal(fraction, number_decimal):
            Convert a fraction into his decimal with an specific amount of
            decimals.
        to_mixed_number(fraction): Convert a fraction into his mixed number.
        to_fraction(number): Convert a decimal into his fraction.
    """

    @classmethod
    def to_decimal(cls, fraction, number_decimal=None):
        """Convert a fraction into his decimal with an specific amount of
        decimals.

        Args:
            fraction (str): Fraction to convert to its decimal
            number_decimal (int, optional): Amount decimal places of the
            decimal. Defaults to None.

        Returns:
            float: Decimal of the fraction
        """

        # Discard invalid argument types
        if not isinstance(fraction, str):  # First argument
            print("The first argument isn't a fraction")
            return None

        elif (
                not number_decimal is None
                and not isinstance(number_decimal, int)
            ):  # Second Argument
            print("The second argument isn't an integer")
            return None

        try:
            numerator, denominator = cls.separate_numbers(fraction)
            decimal = numerator / denominator  # Calculate decimal

            if number_decimal is not None:  # Optional argument
                decimal = round(decimal, number_decimal)

            return decimal

        except TypeError:  # Discard invalid string arguments
            print("One of the arguments isn't a valid argument")

    @classmethod
    def to_mixed_number(cls, fraction):
        """Convert a fraction into his mixed number.

        Args:
            fraction (str): Fraction to convert to a mixed number

        Returns:
            str: Mixed number of the fraction
        """

        if not isinstance(fraction, str):  # Discard invalid argument types
            print("The argument isn't a fraction")
            return None

        try:
            numerator, denominator = cls.separate_numbers(fraction)

            if denominator == 0:  # ZeroDivision Error
                print("Cannot be divided by zero")
                return None

            if abs(numerator) > abs(denominator):
                # Calculate integer part
                if 1 > (numerator / denominator) > 0:  # If trunc(x) = None
                    integer = 1

                else:  # Other cases
                    integer = int(trunc(numerator / denominator))

                # Calculate new numerator
                numerator = abs(numerator) % abs(denominator)

                if numerator != 0:
                    fraction = str(numerator) + "/" + str(abs(denominator))
                else:  # If res = 0 so there is a perfect division
                    return integer

                fraction = cls.simplify(fraction)
                mixed_number = str(integer) + " " + fraction
                return mixed_number

            return fraction

        except TypeError:  # Discard invalid string arguments
            print("The argument isn't valid")

    @classmethod
    def to_fraction(cls, number):
        """Convert a decimal into his fraction.

        Args:
            number (int/float): Number to convert to a fraction

        Returns:
            str: Fraction of the number
        """

        try:
            if isinstance(number, str):  # Case mixed number
                # Separate integer and fraction part
                numbers = number.split(" ")
                numbers[1] = cls.sign_checker(numbers[1])
                numerator, denominator = cls.separate_numbers(numbers[1])

                # Check if is negative or positive
                # Two parts in negative
                if "-" in numbers[0] and "-" in str(numerator):
                    integer = abs(int(numbers[0]))
                    numerator = (integer * denominator + abs(numerator))

                # One part in negative
                elif "-" in numbers[0] or "-" in str(numerator):
                    integer = abs(int(numbers[0]))
                    numerator = (integer * denominator + abs(numerator)) * -1

                else:  # Zero parts in negative
                    integer = int(numbers[0])
                    numerator = integer * denominator + numerator

                fraction = str(numerator) + "/" + str(denominator)
                fraction = cls.simplify(fraction)

            elif isinstance(number, (int, float)):
                # Case int and float
                # Separate integer and decimal
                numbers = str(float(number)).split(".")
                # Count number of decimals
                decimal_amount = len(numbers[1])
                # Calculate numerator
                numerator = round(float(number) * pow(10, decimal_amount), 0)
                numerator = int(numerator)
                # Calculate denominator
                denominator = pow(10, decimal_amount)
                # New fraction
                fraction = str(numerator) + "/" + str(denominator)
                fraction = cls.simplify(fraction)

            else:
                # Discard invalid argument types
                print("The argument isn't valid")
                return None

            return fraction

        except (TypeError, IndexError):  # Discard invalid string arguments
            print("The argument isn't valid")
